/****************************************************************
 * False Leveler
 * Artificial Histogram Matching
 * Jared Sylvester, 2013
 *
 * www.jsylvest.com | jared@jsylvest.com | @jsylvest
 * vimeo.com/jsylvest/videos
 *
 * Released under GPL v3 License.
 *   Feel free to use and modify this code as you wish
 *   but be a mensch and throw a little credit my way
 *
 * This code is very much still a work in progress.
 *   I take no liability, implicit or otherwise, etc. etc.
 ****************************************************************/

/* For information on the generalized Kw distribution:
 http://arxiv.org/pdf/1004.0911v1.pdf
 NB: 4-param version with closed form CDF & inverse CDF */

String inputImage = "carnival01b.jpg";

final int DISTRO_UNIFORM   = 1001;
final int DISTRO_KUMARA    = 1002;
final int DISTRO_GKUMARA4P = 1003;
int distro = DISTRO_GKUMARA4P;

boolean saveFrames = true;

boolean useLAB = false;

PImage img;

int histSize;
float[] srcHistR;
float[] srcHistG;
float[] srcHistB;
float[] srcHist;
int n;

float[] srcHistL;
float[] srcHistA;

float[] aParam = new float[4];
float[] bParam = new float[4];
float[] cParam = new float[4];
float[] dParam = new float[4];
float[] aOff = new float[4];
float[] bOff = new float[4];
float[] cOff = new float[4];
float[] dOff = new float[4];
float[] aStep = new float[4];
float[] bStep = new float[4];
float[] cStep = new float[4];
float[] dStep = new float[4];

boolean showHisto = false;
boolean colorHisto = true;

boolean bwSrc = false;
boolean bwNewMethod = true;

boolean smoothPDFs = true;

float[][] memo;

boolean showOrigHisto = false;
boolean showDestHisto = false;
boolean showOrigPDF =  true;
boolean showDestPDF =  true;

double[][] srcPDF;

float minVal =    0;
float maxVal =  110;

void setup() {
  //noiseSeed(0);
  //randomSeed(0);
  img = loadImage(inputImage);
  size(img.width, img.height);
  n = img.width*img.height;
  frameRate(60);
  //noLoop();
  noiseDetail(5);
  ellipseMode(RADIUS);
  if ( !useLAB ) {
    histSize = 256;
    srcHistR = new float[histSize];
    srcHistG = new float[histSize];
    srcHistB = new float[histSize];
    srcHist = new float[histSize];
  } 
  else {
    histSize = 300;
    srcHistL = new float[histSize];
    srcHistA = new float[histSize];
    srcHistB = new float[histSize];
  }
  memo = new float[histSize][4];
  println(nfc(n)+" pixels");
  loadPixels();
  img.loadPixels();
  float r = 0, g = 0, b = 0, v = 0;
  srcPDF = new double[histSize][3];

  /*
  float[] lab = new float[3];
   float minL =  10000;
   float maxL = -10000;
   float minA =  10000;
   float maxA = -10000;
   float minB =  10000;
   float maxB = -10000;
   for( int i = 0; i < n; i++ ) {
   r = red(img.pixels[i]);
   g = green(img.pixels[i]);
   b = blue(img.pixels[i]);
   lab = rgb2xyz(r,g,b);
   ///print(r+", "+g+", "+b+";  ");
   ///println(lab[0]+", "+lab[1]+", "+lab[2]);
   if( lab[0] > maxL ) maxL = lab[0];
   if( lab[0] < minL ) minL = lab[0];
   if( lab[1] > maxA ) maxA = lab[1];
   if( lab[1] < minA ) minA = lab[1];
   if( lab[2] > maxB ) maxB = lab[2];
   if( lab[2] < minB ) minB = lab[2];
   }
   println("minL = "+nfs(minL,3,2)+";  maxL = "+nfs(maxL,3,2));
   println("minA = "+nfs(minA,3,2)+";  maxA = "+nfs(maxA,3,2));
   println("minB = "+nfs(minB,3,2)+";  maxB = "+nfs(maxB,3,2));
   */

  if ( useLAB ) {
    float[] lab = new float[3];
    int labL, labA, labB;
    for ( int i = 0; i < n; i++ ) {
      r = red(img.pixels[i]);
      g = green(img.pixels[i]);
      b = blue(img.pixels[i]);
      lab = rgb2xyz(r, g, b);
      labL = round(map(lab[0], minVal, maxVal-12, 0, histSize));
      labA = round(map(lab[1], minVal, maxVal-8, 0, histSize));
      labB = round(map(lab[2], minVal, maxVal, 0, histSize));
      for ( int j = labL; j < histSize; j++ ) { 
        srcHistL[j] += 1;
      }
      for ( int j = labA; j < histSize; j++ ) { 
        srcHistA[j] += 1;
      }
      for ( int j = labB; j < histSize; j++ ) { 
        srcHistB[j] += 1;
      }
      srcPDF[labL][0] += 1;
      srcPDF[labA][1] += 1;
      srcPDF[labB][2] += 1;
      if ( smoothPDFs ) {
        srcPDF[constrain(round(labL-1), 0, histSize-1)][0] += 1.0;
        srcPDF[constrain(round(labL+1), 0, histSize-1)][0] += 1.0;
        srcPDF[constrain(round(labA-1), 0, histSize-1)][1] += 1.0;
        srcPDF[constrain(round(labA+1), 0, histSize-1)][1] += 1.0;
        srcPDF[constrain(round(labB-1), 0, histSize-1)][2] += 1.0;
        srcPDF[constrain(round(labB+1), 0, histSize-1)][2] += 1.0;
      }
    }
  } 
  else {

    for ( int i = 0; i < n; i++ ) {
      r = red(img.pixels[i]);
      g = green(img.pixels[i]);
      b = blue(img.pixels[i]);
      if ( bwNewMethod && bwSrc ) {
        r = red(colormap(red(img.pixels[i])/255.0));
        g = green(colormap(green(img.pixels[i])/255.0));
        b = blue(colormap(blue(img.pixels[i])/255.0));
      }
      for ( int j = int(r); j < histSize; j++ ) {
        srcHistR[j] += 1;
      }
      for ( int j = int(g); j < histSize; j++ ) {
        srcHistG[j] += 1;
      }
      for ( int j = int(b); j < histSize; j++ ) {
        srcHistB[j] += 1;
      }
      v = round((r+g+b)/3.0);
      for ( int j = int(v); j < histSize; j++ ) {
        srcHist[j] += 1;
      }
      srcPDF[round(r)][0] += 1;
      srcPDF[round(g)][1] += 1;
      srcPDF[round(b)][2] += 1;
      if ( smoothPDFs ) {
        srcPDF[constrain(round(r-1), 0, histSize-1)][0] += 1.0;
        srcPDF[constrain(round(r+1), 0, histSize-1)][0] += 1.0;
        srcPDF[constrain(round(g-1), 0, histSize-1)][1] += 1.0;
        srcPDF[constrain(round(g+1), 0, histSize-1)][1] += 1.0;
        srcPDF[constrain(round(b-1), 0, histSize-1)][2] += 1.0;
        srcPDF[constrain(round(b+1), 0, histSize-1)][2] += 1.0;
      }
    }
  }

  double divisor = (double)(n);
  if ( smoothPDFs ) {
    divisor = (double)(3.0*n);
  }
  for ( int i = 0; i < histSize; i++ ) {
    if ( useLAB ) {
      srcHistL[i] = srcHistL[i] / float(n);
      srcHistA[i] = srcHistL[i] / float(n);
      srcHistB[i] = srcHistL[i] / float(n);
    } 
    else {
      srcHistR[i] = srcHistR[i] / float(n);
      srcHistG[i] = srcHistG[i] / float(n);
      srcHistB[i] = srcHistB[i] / float(n);
      srcHist[i]  = srcHist[i]  / float(n);
    }
    srcPDF[i][0] = srcPDF[i][0] / divisor;
    srcPDF[i][1] = srcPDF[i][1] / divisor;
    srcPDF[i][2] = srcPDF[i][2] / divisor;
  }
  for ( int i = 0; i < 4; i++ ) {
    aParam[i] = 1.0;
    bParam[i] = 1.0;
    cParam[i] = 1.0;
    dParam[i] = 1.0;
    aOff[i] = random(0.0, 1000.0);
    bOff[i] = random(0.0, 1000.0);
    cOff[i] = random(0.0, 1000.0);
    dOff[i] = random(0.0, 1000.0);
    /* increase these values to make colors shift faster & vice-versa */
    aStep[i] = 0.005;
    bStep[i] = 0.005; 
    cStep[i] = 0.005; 
    dStep[i] = 0.005;
  }
}

void draw() {
  update();
  float v1, v2;
  float r1 = 0, g1 = 0, b1 = 0;
  float r2 = 0, g2 = 0, b2 = 0;
  int labL1, labA1, labB1;
  float labL2, labA2, labB2;
  float[] lab = new float[3];
  float[] rgb = new float[3];

  color c1 = color(0, 0, 0);
  color c2 = color(0, 0, 0);
  for ( int i = 0; i < histSize; i++ ) {
    memo[i][0] = -1;
    memo[i][1] = -1;
    memo[i][2] = -1;
    memo[i][3] = -1;
  }
  for ( int x = 0; x < width; x++ ) {
    for ( int y = 0; y < height; y++ ) {
      c1 = img.pixels[y*width+x];
      if ( !useLAB ) {
        if ( !bwNewMethod && bwSrc ) {
          v1 = (red(c1)+green(c1)+blue(c1))/3.0;
          if ( memo[round(v1)][4] == -1 ) {
            v2 = transform(v1, srcHist, aParam[3], bParam[3], cParam[3], dParam[3]);
            memo[round(v1)][4] = v2;
          } 
          else {
            v2 = memo[round(v1)][4];
          }
          c2 = color(v2);
        } 
        else {
          if ( bwSrc ) {
            c1 = colormap((red(c1)+green(c1)+blue(c1))/(3.0*255.0));
          }
          r1 = red(c1);   // c1 >> 16 & 0xFF
          g1 = green(c1); // c1 >> 8 & 0xFF
          b1 = blue(c1);  // c1 & 0xFF

          if ( memo[round(r1)][0] == -1 ) {
            r2 = transform(r1, srcHistR, aParam[0], bParam[0], cParam[0], dParam[0]);
            memo[round(r1)][0] = r2;
          } 
          else {
            r2 = memo[round(r1)][0];
          }

          if ( memo[round(g1)][1] == -1 ) {
            g2 = transform(g1, srcHistG, aParam[1], bParam[1], cParam[1], dParam[1]);
            memo[round(g1)][1] = g2;
          } 
          else {
            g2 = memo[round(g1)][1];
          }

          if ( memo[round(b1)][2] == -1 ) {
            b2 = transform(b1, srcHistB, aParam[2], bParam[2], cParam[2], dParam[2]);
            memo[round(b1)][2] = b2;
          } 
          else {
            b2 = memo[round(b1)][2];
          }

          c2 = color(r2, g2, b2);
        }
      } 
      else {
        lab = rgb2xyz( red(c1), green(c1), blue(c1) );
        labL1 = round(map(lab[0], minVal, maxVal-12, 0, histSize));
        labA1 = round(map(lab[1], minVal, maxVal-8, 0, histSize));
        labB1 = round(map(lab[2], minVal, maxVal, 0, histSize));
        if ( memo[labL1][0] == -1 ) {
          labL2 = transform(labL1, srcHistL, aParam[0], bParam[0], cParam[0], dParam[0]);
          memo[labL1][0] = labL2;
        } 
        else {
          labL2 = memo[labL1][0];
        }
        if ( memo[labA1][1] == -1 ) {
          labA2 = transform(labA1, srcHistL, aParam[0], bParam[0], cParam[0], dParam[0]);
          memo[labA1][1] = labA2;
        } 
        else {
          labA2 = memo[labA1][1];
        }
        if ( memo[labB1][2] == -1 ) {
          labB2 = transform(labB1, srcHistL, aParam[0], bParam[0], cParam[0], dParam[0]);
          memo[labB1][2] = labB2;
        } 
        else {
          labB2 = memo[labB1][2];
        }
        rgb = xyz2rgb(labL2, labA2, labB2);
        c2 = color(rgb[0], rgb[1], rgb[2]);
      }
      pixels[y*width+x] = c2;
    }
  }
  if (mousePressed == true) {
    image(img, 0, 0);
  } 
  else {
    updatePixels();
  }
  ///println(red(c1)+", "+green(c1)+", "+blue(c1));
  ///println(srcHistR[round(red(c1))]);

  if ( saveFrames ) {
    saveFrame("out/frame-####.jpg");
  }
  if ( showHisto ) {
    if ( showOrigHisto ) { 
      plotOrigHisto();
    }
    if ( showDestHisto ) { 
      plotDestHisto();
    }
    if ( showOrigPDF ) { 
      plotOrigPDF();
    }
    if ( showDestPDF ) { 
      plotDestPDF();
    }
  }
  fill(0);
  text(nf(frameRate, 2, 1), 10, height-20);
}

float transform(float src, float[] hist, float a, float b, float c, float d) {
  float result = hist[round(src)];
  if ( distro == DISTRO_UNIFORM ) {
    // do nothing
  } 
  else if ( distro == DISTRO_KUMARA ) {
    result = pow( 1.0 - pow(1.0-result, b), a );
  } 
  else if ( distro == DISTRO_GKUMARA4P ) {
    result = pow(1.0-pow(1.0-pow(1.0-pow(1.0-result, 1.0/(d+1)), c), b), a);
  }
  result = result * 255.0;
  return result;
}

/*
color c0 = color(  0,   0, 255);
 color c1 = color(128, 128, 128);
 color c2 = color(255,   0,   0);
 */

// http://www.sandia.gov/~kmorel/documents/ColorMaps/
color c0 = color( 0.217*255, 0.525*255, 0.910*255 );
color c1 = color( 0.865*255, 0.865*255, 0.865*255 );
color c2 = color( 0.677*255, 0.492*255, 0.093*255 );

color colormap(float src) {
  color res;
  if ( src < 0.5 ) {
    res = lerpColor(c0, c1, map(src, 0.0, 0.5, 0.0, 1.0));
  } 
  else {
    res = lerpColor(c1, c2, map(src, 0.5, 1.0, 0.0, 1.0));
  }
  return res;
}

void update() {
  if ( distro == DISTRO_KUMARA ) {
    float expon = 4.0;
    for ( int i = 0; i < 4; i++ ) {
      aOff[i] = aOff[i] + aStep[i];
      bOff[i] = bOff[i] + bStep[i];
      if ( !useLAB ) {
        aParam[i] = pow(2.0, expon*( noise(aOff[i]) - 0.5 ));
        bParam[i] = pow(2.0, expon*( noise(bOff[i]) - 0.5 ));
      } 
      else {
        aParam[i] = 1.0+noise(aOff[i]);
        bParam[i] = pow(2.0, 8.0*( noise(bOff[i]) - 0.75 ));
        /*
        aParam[i] = 1.0/map(noise(aOff[i]), 0.0, 1.0, 1.7, 2.3);
         bParam[i] = 1.0/map(noise(bOff[i]), 0.0, 1.0, 1.7, 6.3);
         */
      }
    }
  }
  if ( distro == DISTRO_GKUMARA4P ) {
    /* increase expon to allow more extreme distribution shapes */
    float expon = 4.0;
    for ( int i = 0; i < 4; i++ ) {
      aOff[i] = aOff[i] + aStep[i];
      bOff[i] = bOff[i] + bStep[i];
      cOff[i] = cOff[i] + cStep[i];
      dOff[i] = dOff[i] + dStep[i];
      aParam[i] = pow(2.0, expon*( noise(aOff[i]) - 0.5 ));
      bParam[i] = pow(2.0, expon*( noise(bOff[i]) - 0.5 ));
      cParam[i] = pow(2.0, expon*( noise(cOff[i]) - 0.5 ));
      dParam[i] = pow(2.0, expon*( noise(dOff[i]) - 0.5 ));
    }
  }
}

void keyPressed() {
  if ( key == 't' ) {
    showHisto = !showHisto;
  } 
  else if ( key == 'T' ) {
    colorHisto = !colorHisto;
  }
  else if ( key == 'y' ) {
    showDestHisto = !showDestHisto;
    showOrigHisto = !showOrigHisto;
    showDestPDF = !showDestPDF;
    showOrigPDF = !showOrigPDF;
  } 
}

