
float[] rgb2xyz( float R, float G, float B ) {
  // R, G, B all in 0 to 255

  float r = R/255.0;
  float g = G/255.0;
  float b = B/255.0;

  if (r > 0.04045) { 
    r = pow((r + 0.055) / 1.055, 2.4);
  }
  else { 
    r = r / 12.92;
  }
  if ( g > 0.04045) { 
    g = pow((g + 0.055) / 1.055, 2.4);
  }
  else { 
    g = g / 12.92;
  }
  if (b > 0.04045) { 
    b = pow((b + 0.055) / 1.055, 2.4);
  }
  else {  
    b = b / 12.92;
  }

  r = r * 100.0;
  g = g * 100.0;
  b = b * 100.0;

  //Observer. = 2°, Illuminant = D65
  float xyz[] = new float[3];
  xyz[0] = r * 0.4124 + g * 0.3576 + b * 0.1805;
  xyz[1] = r * 0.2126 + g * 0.7152 + b * 0.0722;
  xyz[2] = r * 0.0193 + g * 0.1192 + b * 0.9505;
  return xyz;
}

float[] xyz2lab( float X, float Y, float Z ) {
  final float REF_X = 95.047; // Observer= 2°, Illuminant= D65
  final float REF_Y = 100.000; 
  final float REF_Z = 108.883;
  
  float x = X / REF_X;
  float y = Y / REF_Y;
  float z = Z / REF_Z;
  
  float oneThird = 1.0/3.0;
  if ( x > 0.008856 ) { x = pow( x , oneThird ); }
  else { x = ( 7.787 * x ) + ( 16/116 ); }
  if ( y > 0.008856 ) { y = pow( y , oneThird ); }
  else { y = ( 7.787 * y ) + ( 16/116 ); }
  if ( z > 0.008856 ) { z = pow( z , oneThird ); }
  else { z = ( 7.787 * z ) + ( 16/116 ); }
  
  float lab[] = new float[3];
  lab[0] = ( 116.0 * y ) - 16.0;
  lab[1] = 500.0 * ( x - y );
  lab[2] = 200.0 * ( y - z );
  
  return lab;
}

float[] lab2xyz( float l, float a, float b ) {
  final float REF_X = 95.047; // Observer= 2°, Illuminant= D65
  final float REF_Y = 100.000; 
  final float REF_Z = 108.883;
  
  float y = (l + 16.0) / 116.0;
  float x = a / 500.0 + y;
  float z = y - b / 200.0;
  
  if ( pow( y , 3 ) > 0.008856 ) { y = pow( y , 3 ); }
  else { y = ( y - 16.0 / 116.0 ) / 7.787; }
  if ( pow( x , 3 ) > 0.008856 ) { x = pow( x , 3 ); }
  else { x = ( x - 16.0 / 116.0 ) / 7.787; }
  if ( pow( z , 3 ) > 0.008856 ) { z = pow( z , 3 ); }
  else { z = ( z - 16.0 / 116.0 ) / 7.787; }
 
  float[] xyz = new float[3];
  xyz[0] = REF_X * x;     
  xyz[1] = REF_Y * y; 
  xyz[2] = REF_Z * z; 
 
  return xyz;
}

float[] xyz2rgb( float X, float Y, float Z ) {
  //X from 0 to  95.047    (Observer = 2°, Illuminant = D65)
  //Y from 0 to 100.000
  //Z from 0 to 108.883
    
  float x = X / 100.0;      
  float y = Y / 100.0;      
  float z = Z / 100.0; 
  
  float r = x * 3.2406 + y * -1.5372 + z * -0.4986;
  float g = x * -0.9689 + y * 1.8758 + z * 0.0415;
  float b = x * 0.0557 + y * -0.2040 + z * 1.0570;
  
  if ( r > 0.0031308 ) { r = 1.055 * pow( r , ( 1.0 / 2.4 ) ) - 0.055; }
  else { r = 12.92 * r; }
  if ( g > 0.0031308 ) { g = 1.055 * pow( g , ( 1.0 / 2.4 ) ) - 0.055; }
  else { g = 12.92 * g; }
  if ( b > 0.0031308 ) { b = 1.055 * pow( b , ( 1.0 / 2.4 ) ) - 0.055; }
  else { b = 12.92 * b; }
  
  float[] rgb = new float[3];
  rgb[0] = round( r*255.0 );
  rgb[1] = round( g*255.0 );
  rgb[2] = round( b*255.0 );
  
  return rgb;
}

float[] rgb2lab( float R, float G, float B ) {
  float[] xyz = new float[3];
  xyz = rgb2xyz(R,G,B);
  float[] lab = new float[3];
  lab = xyz2lab(xyz[0], xyz[1], xyz[2]);
  return lab;
}

float[] lab2rgb( float l, float a, float b ) {
  float[] xyz = new float[3];
  xyz = lab2xyz(l,a,b);
  float[] rgb = new float[3];
  rgb = xyz2rgb(xyz[0], xyz[1], xyz[2]);
  return rgb;
}
